import unittest


class Test(unittest.TestCase):
    """Unit tests for seqtools."""

    def test_load_fasta_file(self):
        pass

    def test_load_seq_string(self):
        pass

    def test_write_fasta_file(self):
        pass

if __name__ == "__main__":
    unittest.main()
